# ----------------------------------------------------------------------
# Copyright (c) 2022
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

# ---------------------
# Third party libraries
# ---------------------


SERIAL_PORT_PREFIX = "/dev/ttyUSB"

# TESS-W data

TEST_IP = "192.168.4.1"
TEST_TCP_PORT = 23
TEST_UDP_PORT = 2255
TEST_SERIAL_PORT = 0
TEST_BAUD = 9600

# Timestamp format
TSTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

# Condensed timestamp
TSTAMP_SESSION_FMT = "%Y-%m-%dT%H:%M:%S"

# ZP of stars3
ZP_ABS = 20.44


type FreqSequence = list[float, ...]
type TimeSequence = list[float, ...]
type NameSequence = list[str, ...]
